﻿/*
 *         Progetto: Problema dei lettori/scrittori (seconda soluzione)
 * 
 *           Autore: Roberto Fuligni
 *  Ultima modifica: 7/3/2019
 *  
 *      Descrizione: In questa soluzione gli scrittori hanno priorità sugi lettori.
 *                   (Rischio di starvation per i lettori).
 *  
 */

using System;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;

namespace LS_Soluzione2
{
    class Program
    {
        const int MaxProcessi = 100;        // Metà lettori, metà scrittori
        const int RitardoLettura = 20;      // ms
        const int RitardoScrittura = 10;    // ms

        static SemaphoreSlim mutexL = new SemaphoreSlim(1);     // Regola l'accesso alla variabile NumLettori
        static SemaphoreSlim mutexS = new SemaphoreSlim(1);     // Regola l'accesso alla variabile NumScrittori
        static SemaphoreSlim sincro = new SemaphoreSlim(1);     // Regola l'accesso al buffer per sacrivere/leggere

        static SemaphoreSlim semLettura = new SemaphoreSlim(1);

        static int numLettori = 0;
        static int numScrittori = 0;
        static int buffer = 100;

        static void InizioLettura()
        {
            semLettura.Wait();  // Indica un nuovo accesso in lettura
            mutexL.Wait();
            numLettori++;
            if (numLettori == 1)
                sincro.Wait();
            mutexL.Release();
            semLettura.Release();
        }

        static void FineLettura()
        {
            mutexL.Wait();
            numLettori--;
            if (numLettori == 0)
                sincro.Release();
            mutexL.Release();
        }

        static void InizioScrittura()
        {
            mutexS.Wait();
            numScrittori++;
            if (numScrittori == 1)
                semLettura.Wait();
            mutexS.Release();

            sincro.Wait();
        }
        static void FineScrittura()
        {
            sincro.Release();
            mutexS.Wait();
            numScrittori--;
            if (numScrittori == 0)
                semLettura.Release();
            mutexS.Release();
        }

        static void Leggi(int idLettore)
        {
            int dato;
            InizioLettura();

            if (RitardoLettura > 0)
                Thread.Sleep(RitardoLettura);
            dato = buffer;
            Console.WriteLine("Il lettore n. {0} ha letto il buffer. Valore rilevato: {1}", idLettore, dato);

            FineLettura();
        }
        static void Scrivi(int idScrittore)
        {
            int dato;
            InizioScrittura();

            if (RitardoScrittura > 0)
                Thread.Sleep(RitardoScrittura);
            dato = buffer;
            dato++;
            buffer = dato;
            Console.WriteLine("Lo scrittore n. {0} ha modificato il buffer. Nuovo valore: {1}", idScrittore, dato);

            FineScrittura();
        }
        static void Main(string[] args)
        {
            List<Task> lista = new List<Task>();
            for (int i = 1; i <= MaxProcessi / 2; i++)
            {
                int id = i;
                // Avvia un nuovo task scrittore
                lista.Add(Task.Factory.StartNew(() => Scrivi(id), TaskCreationOptions.LongRunning));
                // Avvia un nuovo task lettore
                lista.Add(Task.Factory.StartNew(() => Leggi(id), TaskCreationOptions.LongRunning));
            }

            // Attende la fine di tutti i lettori/scrittori 
            Task.WaitAll(lista.ToArray());
        }
    }
}
